EXTLD(NeuralRack_png)
EXTLD(approved_png)
EXTLD(arrow_png)
EXTLD(cancel_png)
EXTLD(choice_png)
EXTLD(colors_png)
EXTLD(directory_png)
EXTLD(directory_open_png)
EXTLD(directory_select_png)
EXTLD(eject_png)
EXTLD(error_png)
EXTLD(exit_png)
EXTLD(exit__png)
EXTLD(file_png)
EXTLD(gear_png)
EXTLD(grid_png)
EXTLD(image_directory_png)
EXTLD(info_png)
EXTLD(menu_png)
EXTLD(message_png)
EXTLD(midikeyboard_png)
EXTLD(norm_png)
EXTLD(question_png)
EXTLD(save_png)
EXTLD(settings_png)
EXTLD(warning_png)
EXTLD(xputty_logo_png)
EXTLD_FONT(RobotoCondensedBold_ttf)
EXTLD_FONT(RobotoCondensedRegular_ttf)
