#pragma once

namespace math_approx
{
}

#include "src/basic_math.hpp"

#include "src/trig_approx.hpp"
#include "src/inverse_trig_approx.hpp"
#include "src/pow_approx.hpp"
#include "src/log_approx.hpp"
#include "src/hyperbolic_trig_approx.hpp"
#include "src/inverse_hyperbolic_trig_approx.hpp"
#include "src/sigmoid_approx.hpp"
#include "src/wright_omega_approx.hpp"
#include "src/polylogarithm_approx.hpp"
