/*
 * NeuralRackClap.cpp
 *
 * Binds the generic CLAP wrapper to the NeuralRack PluginClient.
 *
 * SPDX-License-Identifier: BSD-3-Clause
 * 
 * Copyright (C) 2025 brummer <brummer@web.de>
 */

#define CLAPIPLUG

#include "PluginClient.cc"
#include "CLAPWrapper.cpp"
