#include "RTNeural.h"

// forward declare some template classes
template class RTNEURAL_NAMESPACE::Model<float>;
template class RTNEURAL_NAMESPACE::Model<double>;
template class RTNEURAL_NAMESPACE::Layer<float>;
template class RTNEURAL_NAMESPACE::Layer<double>;
